-- Quasi.adb	Quasi-Random number generator package implementation

--    (c) Copyright 1997 Everett F. Carter Jr.   Permission is granted
--    by the author to use this software for any application provided
--    this copyright notice is preserved.

-- $Author:   skip  $
-- $Workfile:   quasirandom.adb  $
-- $Revision:   1.0  $
-- $Date:   24 Jul 1997 11:19:30  $

-- ==================================================================

with Vector; use Vector;
with BITOPS; use BITOPS;	-- get bitwise operators

package body QuasiRandom is

        maxbit : constant := 30;
	mdim   : constant := maxdim-1;
	mbit   : constant := maxbit-1;
                
	ip    : array(0..mdim) of INTEGER;
        mdeg  : array(0..mdim) of INTEGER;
        iv    : array(0..1560) of INTEGER;

        factor : float;

	function Index(k : integer; j : integer) return integer is
            begin
              return k + (j-1)*maxdim;
            end Index;

	function Dimension(qr : QRrec) return integer is
           begin
             return qr.dim;
           end Dimension;

        function Status(qr : QRrec) return integer is
           begin
             return qr.err_flag;
           end Status;
           
	procedure Initialize(qr : in out QRrec; dim : integer) is
           begin

             if dim > maxdim then raise CONSTRAINT_ERROR; end if;

	     qr.dim := dim;
             qr.index := 0;
	     qr.err_flag := 0;
             
	     for k in 0..dim loop
             		qr.ix(k) := 0;
             end loop;
                          
           end Initialize;


	procedure Number(qr : in out QRrec; x : out Vector_Type) is
           i,im     : integer;
           begin
	     -- find rightmost zero bit
             im := qr.index;
             qr.index := qr.index + 1;

             for j in 0..mbit loop
                 if BIT_AND(im, 1) = 0 then
                        i := j * maxdim;
                        exit;
                 end if;
                 im := im / 2;
             end loop;

	     for k in 0..qr.dim-1 loop
             	 qr.ix(k) := BIT_XOR( qr.ix(k), iv(i+k)  );
                 x(k)  := float( qr.ix(k) ) * factor;
             end loop;

           end number;

	procedure PackageInitialize is		-- called once

           limit : constant := 2**29;
           
           niv, mval, i, ipp, k : integer;
           begin

	     -- the primitive polynomial coefficients for up to degree 8
             ip := ( 0, 1, 1, 2, 1, 4, 2, 4, 7, 11, 13, 14,
			  1, 13, 16, 19, 22, 25,
                          1, 4, 7, 8, 14, 19, 21, 28, 31, 32, 37, 41, 42,
                          50, 55, 56, 59, 62,
                          14, 21, 22, 38, 47, 49, 50, 52, 56, 67, 70, 84,
                          97, 103, 115, 122 );

              -- the degree of each polynomial
	      mdeg := ( 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5,
			    6, 6, 6, 6, 6, 6,
                            7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
                            7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8,
                            8, 8, 8, 8, 8, 8, 8, 8, 8, 8 );


	     niv := maxdim * maxbit - 1;

             for j in 0..niv loop
             		iv(j) := 0;
             end loop;

             for j in 0..mdim loop
             		iv(j) := 1;
             end loop;

             mval := 4;		ipp := 1;

	     k := maxdim;	i := 0;
             while k < niv loop
		iv(k) := ipp;
		i := i + 1;
                if i = maxdim then
                	if mval < limit then
                        	mval := 2 * mval;
                        end if;
                        ipp  := ipp + 1;
                        i := 0;
                end if;

                if ipp > mval then
                	ipp := 1;
                end if;

                iv(k+1) := ipp;

                i := i + 1;
                if i = maxdim then
                	if mval < limit then
                		mval := 2 * mval;
                        end if;
                        ipp  := ipp + 2;
                        i := 0;
                else
                	ipp := ipp + 2;
                        if ipp > mval then
                        	ipp := 1;
                        end if;
                end if;

                k := k + 2;
             end loop;
             
	     for k in 0..mdim loop

		-- normalize the set iv values
                for j in 1..mdeg(k) loop
                    iv( Index(k,j) ) := iv( Index(k,j) ) * (2 ** (maxbit-j));
                end loop;

                -- calculate the rest of the iv values
		for j in mdeg(k)+1..maxbit loop

		    ipp := ip(k);

                    -- calculate Gray code of iv
                    i := iv( Index(k, j - mdeg(k)) );
                    i := Bit_XOR(i, i / (2 ** mdeg(k)) );

                    for l in reverse 1..mdeg(k)-1 loop
			if Bit_AND(ipp, 1 ) = 1 then
                        		i := Bit_XOR(i, iv(Index(k,j-l)) );
			end if;

                        ipp := ipp / 2;
                        
                    end loop;

                    iv( Index(k,j) ) := i;
                    
                end loop;
                
             end loop;

             factor := 1.0 / (2.0 ** maxbit);
             
           end PackageInitialize;

    begin	-- package initialization

	PackageInitialize;
        
end QuasiRandom;
